@@ -1,5 +1,6 @@
 import collections
 import json
+import logging
 import os.path
 import shutil
 import sys
@@ -11,6 +12,7 @@ from typing import (Any, BinaryIO, Dict, Iterator, List, Optional, Set, TextIO,
 import toposort
 from haoda.backend import xilinx as hls
 
+import tapa.autobridge as autobridge
 from tapa import util
 from tapa.verilog import ast
 from tapa.verilog import xilinx as rtl
@@ -18,6 +20,8 @@ from tapa.verilog import xilinx as rtl
 from .instance import Instance, Port
 from .task import Task
 
+_logger = logging.getLogger().getChild(__name__)
+
 STATE00 = ast.IntConst("2'b00")
 STATE01 = ast.IntConst("2'b01")
 STATE11 = ast.IntConst("2'b11")
@@ -172,7 +176,7 @@ class Program:
 
   def instrument_rtl(
       self,
-      directive: Optional[Tuple[Dict[str, Any], TextIO]] = None,
+      directive: Optional[Dict[str, Any]] = None,
       register_level: int = 0,
   ) -> 'Program':
     """Instrument HDL files generated from HLS.
@@ -184,23 +188,29 @@ class Program:
     Returns:
         Program: Return self.
     """
-    # extract and parse RTL
+    # extract and parse RTL and populate tasks
+    oldcwd = os.getcwd()
+    os.chdir(self.work_dir)
     for task in self._tasks.values():
-      oldcwd = os.getcwd()
-      os.chdir(self.work_dir)
       task.module = rtl.Module([self.get_rtl(task.name)])
-      os.chdir(oldcwd)
+      self._populate_task(task)
+    os.chdir(oldcwd)
 
     # generate partitioning constraints if partitioning directive is given
     if directive is not None:
-      self._process_partition_directive(directive[0], directive[1])
+      floorplan = directive.get('floorplan')
+      if floorplan is None:
+        floorplan = autobridge.generate_floorplan(
+            self.top_task,
+            self.work_dir,
+            self._get_fifo_width,
+            directive['connectivity'],
+            directive['part_num'],
+        )
+      self._process_partition_directive(floorplan, directive['constraint'])
     if register_level:
       self.top_task.module.register_level = register_level
 
-    # populate tasks
-    for task in self._tasks.values():
-      self._populate_task(task)
-
     # instrument the upper-level RTL
     for task in self._tasks.values():
       if task.is_upper:
@@ -250,10 +260,13 @@ class Program:
 
     topology: Dict[str, Dict[str, List[str]]] = {}
 
+    # list of strings that creates the pblocks, may be empty
+    pblock_tcl: List[str] = []
+
     for instance in self.top_task.instances:
       for arg in instance.args:
         if arg.cat == Instance.Arg.Cat.ASYNC_MMAP:
-          instance_dict[rtl.async_mmap_instance_name(arg.name)] = ''
+          instance_dict[rtl.async_mmap_instance_name(arg.mmap_name)] = ''
       instance_dict[instance.name] = ''
 
     # check that all non-FIFO instances are assigned to one and only one SLR
@@ -268,6 +281,7 @@ class Program:
           subset.add(instance_obj)
         else:
           topology[region] = instance_obj
+          pblock_tcl.append(instance_obj.pop('tcl', ''))
       intersect = subset & directive_instance_set
       if intersect:
         raise InputError('more than one region assignment: %s' % intersect)
@@ -309,10 +323,14 @@ class Program:
           instance_dict[rtl.fifo_partition_name(name, idx)] = region
         fifo_partition_count[name] = idx + 1
 
-    rtl.print_constraints(instance_dict, constraints)
+    rtl.print_constraints(instance_dict, constraints, pre=''.join(pblock_tcl))
 
     self.top_task.module.register_level = max(
-        map(len, topology[instance_dict[self.ctrl_instance_name]].values())) + 1
+        map(len, topology[instance_dict[self.ctrl_instance_name]].values()),
+        default=-1,
+    ) + 1
+    _logger.info('top task register level set to %d based on floorplan',
+                 self.top_task.module.register_level)
     self.top_task.module.fifo_partition_count = fifo_partition_count
 
   def _populate_task(self, task: Task) -> None:
@@ -367,15 +385,9 @@ class Program:
         continue
 
       # add FIFO instances
-      task_name, _, fifo_port = task.get_fifo_port(fifo_name, 'produced_by')
-      child_ports = self.get_task(task_name).module.ports
-      width = child_ports[rtl.fifo_port_name(fifo_port,
-                                             rtl.OSTREAM_SUFFIXES[0])].width
-      # TODO: err properly if not integer literals
-      fifo_width = int(width.msb.value) - int(width.lsb.value) + 1
       task.module.add_fifo_instance(
           name=fifo_name,
-          width=fifo_width,
+          width=self._get_fifo_width(task, fifo_name),
           depth=fifo['depth'],
       )
 
@@ -720,3 +732,10 @@ class Program:
 
     task.module.add_pipeline(self.start_q, init=rtl.START)
     task.module.add_pipeline(self.done_q, init=is_state(STATE10))
+
+  def _get_fifo_width(self, task: Task, fifo: str) -> int:
+    producer_task, _, fifo_port = task.get_fifo_port(fifo, 'produced_by')
+    fifo_data_port = rtl.fifo_port_name(fifo_port, rtl.OSTREAM_SUFFIXES[0])
+    width = self.get_task(producer_task).module.ports[fifo_data_port].width
+    # TODO: err properly if not integer literals
+    return int(width.msb.value) - int(width.lsb.value) + 1
